@@ -39,18 +39,23 @@
  * If the string ends with a trailing :, letters following are taken as game
  * options. At present, only 'p' (equivalent to an initial 'p' command) is
  * defined.
+ *
+ * SPDX-License-Identifier: BSD-2-Clause
  */
 
 static char *version = "Greed v" RELEASE;
 
 #include <stdlib.h>
+#include <unistd.h>
 #include <string.h>
 #include <ctype.h>
 #include <curses.h>
+#include <term.h>
 #include <signal.h>
 #include <pwd.h>
 #include <fcntl.h>
 #include <stdbool.h>
+#include <time.h>
 #ifdef A_COLOR
 #include <ctype.h>
 #endif
@@ -71,12 +76,17 @@ static char *version = "Greed v" RELEASE;
 
 #define LOCKPATH "/tmp/Greed.lock"	/* lock path for high score file */
 
+#define LOCALSCOREFILE	".greedscores"
+
+#define USERNAMELEN	32	/* length of leading segment that we keep */
+
 /* 
  * changing stuff in this struct
  * makes old score files incompatible
  */
 struct score {
-    char user[9];
+    char user[USERNAMELEN + 1];
+    time_t time;
     int score;
 };
 
@@ -87,6 +97,8 @@ static char *cmdname;
 static WINDOW *helpwin = NULL;
 
 static void topscores(int);
+static int tunnel(chtype, int *);
+static int othermove(int, int);
 
 static void botmsg(char *msg, bool backcur)
 /* 
@@ -158,17 +170,18 @@ static void showscore(void)
  * cursor back on the grid, and refreshes the screen.
  */
 {
-    mvprintw(23, 7, "%d  %.2f%%", score, (float) score / 17.38);
+    mvprintw(23, 7, "%d  %.2f%%",
+	     score,
+	     (float)(score * 100) / (HEIGHT * WIDTH));
     move(y, x);
     refresh();
 }
 
 void showmoves(bool, int*);
 
-main(int argc, char **argv)
+int main(int argc, char **argv)
 {
     int val = 1;
-    extern long time();
     int attribs[9];
 #ifdef A_COLOR
     char *colors;
@@ -219,7 +232,7 @@ main(int argc, char **argv)
 	attribs[8] = COLOR_PAIR(4) | A_BOLD;
 
 	if ((colors = getenv("GREEDOPTS")) != (char *) NULL) {
-	    static char *cnames = " bgcrmywBGCRMYW";
+	    static char *cnames = " rgybmcwRGYBMCW";
 	    char *cp;
 
 	    for (cp = colors; *cp && *cp != ':'; cp++)
@@ -283,7 +296,7 @@ main(int argc, char **argv)
 }
 
 
-int tunnel(chtype cmd, int *attribs)
+static int tunnel(chtype cmd, int *attribs)
 /* 
  * tunnel() does the main game work.  Returns 1 if everything's okay, 0 if
  * user "died", and -1 if user specified and confirmed 'q' (fast quit).
@@ -403,7 +416,7 @@ int tunnel(chtype cmd, int *attribs)
     return (1);
 }
 
-int othermove(int bady, int badx)
+static int othermove(int bady, int badx)
 /* 
  * othermove() checks area for an existing possible move.  bady and
  * badx are direction variables that tell othermove() they are
@@ -484,12 +497,15 @@ void showmoves(bool on, int *attribs)
     }
 }
 
-char doputc(char c)
+static int doputc(int c)
 /* doputc() simply prints out a character to stdout, used by tputs() */
 {
     return(fputc(c, stdout));
 }
 
+/* hack to ignore GCC Unused Result */
+#define IGNORE(r) do{if(r){}}while(0)
+
 static void topscores(int newscore)
 /* 
  * topscores() processes its argument with the high score file, makes any
@@ -503,6 +519,7 @@ static void topscores(int newscore)
     struct score *toplist = (struct score *) malloc(SCOREFILESIZE);
     struct score *ptrtmp, *eof = &toplist[MAXSCORES], *new = NULL;
     extern char *getenv(), *tgetstr();
+    struct passwd *whoami;
     void lockit(bool);
 
     (void) signal(SIGINT, SIG_IGN);	/* Catch all signals, so high */
@@ -510,19 +527,25 @@ static void topscores(int newscore)
     (void) signal(SIGTERM, SIG_IGN);	/* messed up with a kill.     */
     (void) signal(SIGHUP, SIG_IGN);
 
+    whoami = getpwuid(getuid());
+    
     /* following open() creates the file if it doesn't exist
      * already, using secure mode
      */
     if ((fd = open(SCOREFILE, O_RDWR|O_CREAT, 0600)) == -1) {
-	    fprintf(stderr, "%s: %s: Cannot open.\n", cmdname,
-		    SCOREFILE);
-	exit(1);
+	IGNORE(chdir(whoami->pw_dir));
+	if ((fd = open(LOCALSCOREFILE, O_RDWR|O_CREAT, 0600)) == -1) {
+	    fprintf(stderr, "%s: ~/%s: Cannot open.\n", cmdname, LOCALSCOREFILE);
+	    exit(1);
+	}
     }
 
     lockit(true);			/* lock score file */
-    for (ptrtmp=toplist; ptrtmp < eof; ptrtmp++) ptrtmp->score = 0;
     /* initialize scores to 0 */
-    read(fd, toplist, SCOREFILESIZE);	/* read whole score file in at once */
+    for (ptrtmp=toplist; ptrtmp < eof; ptrtmp++)
+	ptrtmp->score = 0;
+    /* read whole score file in at once */
+    IGNORE(read(fd, toplist, SCOREFILESIZE));
 
     if (newscore) {			/* if possible high score */
 	for (ptrtmp=toplist; ptrtmp < eof; ptrtmp++)
@@ -537,18 +560,17 @@ static void topscores(int newscore)
 	    }
 
 	    new->score = newscore;	/* fill "new" with the info */
-	    strncpy(new->user, getpwuid(getuid())->pw_name, 8);
+	    new->time = time(NULL);	/* include a timestamp */
+	    strncpy(new->user, whoami->pw_name, USERNAMELEN);
 	    (void) lseek(fd, 0, 0);	/* seek back to top of file */
-	    write(fd, toplist, SCOREFILESIZE);	/* write it all out */
+	    IGNORE(write(fd, toplist, SCOREFILESIZE));	/* write it all out */
 	}
     }
 
     close(fd);
     lockit(false);			/* unlock score file */
 
-    if (toplist->score) 
-	puts("Rank  Score  Name     Percentage");
-    else 
+    if (!toplist->score) 
 	puts("No high scores.");	/* perhaps "greed -s" was run before *
 					 * any greed had been played? */
     if (new && tgetent(termbuf, getenv("TERM")) > 0) {
@@ -562,10 +584,18 @@ static void topscores(int newscore)
 
     /* print out list to screen, highlighting new score, if any */
     for (ptrtmp=toplist; ptrtmp < eof && ptrtmp->score; ptrtmp++, count++) {
+	struct tm when;
+	char timestr[27];
 	if (ptrtmp == new && boldon)
 	    tputs(boldon, 1, doputc);
-	printf("%-5d %-6d %-8s %.2f%%\n", count, ptrtmp->score,
-	       ptrtmp->user, (float) ptrtmp->score / 17.38);
+	(void)localtime_r(&ptrtmp->time, &when);
+	(void)strftime(timestr, sizeof(timestr), "%Y-%m-%dT%H:%M:%S", &when);
+	printf("%-5d %s %6d %5.2f%% %s\n",
+	       count,
+	       timestr,
+	       ptrtmp->score,
+	       (float) (ptrtmp->score * 100) / (HEIGHT * WIDTH),
+	       ptrtmp->user);
 	if (ptrtmp == new && boldoff) tputs(boldoff, 1, doputc);
     }
 }
@@ -583,7 +613,7 @@ void lockit(bool on)
     int fd, x = 1;
 
     if (on) {
-	while ((fd = creat(LOCKPATH, 0)) == -1) {
+	while ((fd = open(LOCKPATH, O_RDWR | O_CREAT | O_EXCL, 0)) < 0) {
 	    printf("Waiting for scorefile access... %d/15\n", x);
 	    if (x++ >= 15) {
 		puts("Overriding stale lock...");
@@ -612,7 +642,6 @@ void help(void)
 	helpwin = newwin(18, 65, 1, 7);
 #ifdef ACS_URCORNER
 	box(helpwin, ACS_VLINE, ACS_HLINE);	/* print box around info */
-	/* put '+' at corners, looks better */
 	(void) waddch(helpwin, ACS_ULCORNER); 
 	mvwaddch(helpwin, 0, 64, ACS_URCORNER);
 	mvwaddch(helpwin, 17, 0, ACS_LLCORNER); 
@@ -626,14 +655,14 @@ void help(void)
 
 	mvwprintw(helpwin, 1, 2,
 		  "Welcome to %s, by Matthew Day <mday@iconsys.uu.net>.",version);
-	msg(3," The object of Greed is to erase as much of the screen as");
-	msg(4," possible by moving around in a grid of numbers.  To move");
-	msg(5," your cursor, simply use the 'hjklyubn' keys or your number");
+	msg(3, " The object of Greed is to erase as much of the screen as");
+	msg(4, " possible by moving around in a grid of numbers.  To move,");
+	msg(5, " use the arrow keys, your number pad, or one of the letters");
 	mvwprintw(helpwin, 6, 2,
-		  " keypad.  Your location is signified by the '%c' symbol.", ME);
-	msg(7," When you move in a direction, you erase N number of grid");
-	msg(8," squares in that direction, N being the first number in that");
-	msg(9," direction.  Your score reflects the total number of squares");
+	       " 'hjklyubn'. Your location is signified by the '%c' symbol.", ME);
+	msg(7, " When you move in a direction, you erase N number of grid");
+	msg(8, " squares in that direction, N being the first number in that");
+	msg(9, " direction.  Your score reflects the total number of squares");
 	msg(10," eaten.  Greed will not let you make a move that would have");
 	msg(11," placed you off the grid or over a previously eaten square");
 	msg(12," unless no valid moves exist, in which case your game ends.");
